translations = {
    "municipality": "kommune",
    # "university": "universitet",
}

one_to_one = {
    "au": "aarhus universitet",
    "dtu": "danmarks tekniske universitet",
    "dti": "teknologisk institut",
    "sdu": "syddansk universitet (university of southern denmark)",
}

replacements = {
    "alexandra inst.": "alexandra instituttet",
    "university of copenhagen": "københavns universitet",
    "uni. copenhagen": "københavns universitet",
    "technical university of denmark": "danmarks tekniske universitet",
    "aarhus universitet": "aarhus universitet",
    "aarhus university": "aarhus universitet",
    "university of southern denmark": "syddansk universitet (university of southern denmark)",
    "syddansk universitet": "syddansk universitet (university of southern denmark)",
    "sdu,": "syddansk universitet (university of southern denmark)", 
    ", sdu": "syddansk universitet (university of southern denmark)", 
    "(sdu)": "syddansk universitet (university of southern denmark)", 
    "danish regions": "danske regioner",
    "(ouh)": "odense universitets hospital",
    "Odense University Hospital": "odense universitets hospital",
    "rigshospitalet": "rigshospitalet",
    "danish technological institute": "teknologisk institut",
    "roskilde university": "roskilde universitet",
    "aalborg university": "aalborg universitet",
    "uc absalon": "Professionshøjskolen Absalon S/I",
    "steno diabetes center": "steno diabetes center",
    "holbæk hospital": "Holbæk sygehus - Smedelundsgade",

    }

def run_name_handler(virk_navn: str) -> str:
    """Lidt dum funktion, som retter typiske engelske termer til dansk og erstatter bestemte termer. Google Translate kræver billing og vi vil sikkert kunne
    holde os under "limit", men på den korte bane tænker jeg at en tilpasning er fint
    
    På sigt bør replacements dict'en flyttes til en seperat fil, da den har relevans andre steder + potentielt som en tabel i dwh.
    """
    term_upd = virk_navn.lower()
    for uc, r in one_to_one.items():
        if uc == term_upd:
            return r
    for t, r in replacements.items():
            if t in term_upd:
                return r
    for en, da in translations.items():
        term_upd = term_upd.replace(en, da)

    return term_upd


from dataclasses import dataclass
import pandas as pd
import sqlalchemy as sa
import urllib
import keyring

@dataclass
class DBHelper:
    server: str
    trusted: bool = True
    wincred: str = None
    
    def __post_init__(self):
        assert self.trusted or (not self.trusted and self.wincred), "Når du vælger trusted=False, så skal du angive wincred. Navnet på den credential, som du ønsker at anvende"
    
    def connect_engine(self, db: str, echo: bool = True) -> None:
        self.engine = self.__db_connect(db, echo)
        self.insp = sa.inspect(self.engine)

    def close(self) -> None:
        self.engine.dispose()
    
    def hent_data(self, db: str, query: str) -> pd.DataFrame:
        engine = self.connect_engine(db, echo=True)
        with open (query, "r") as q:
            df = pd.read_sql_query(q.read(), engine)
        return df    
    
    def __db_connect(self, db: str, echo: bool) -> sa.Engine:
        """
        Understøtter to muligheder for tilslutning til sql server
        Hvis trusted så anvendes den pågældende windows user, som kører scriptet
        Hvis ikke trused, så anvendes et sæt windows credentials til at logge på
        """
        if self.trusted:
            con_string = (
                f"DRIVER={{SQL Server}};"
                f"SERVER={self.server};"
                f"DATABASE={db};"
                f"trusted_Connection={self.trusted}"
            )
            quoted_con_string = urllib.parse.quote_plus(con_string)
            db_engine = sa.create_engine(
                f"mssql+pyodbc:///?odbc_connect={quoted_con_string}", use_setinputsizes=False, echo=echo
            )
        else:
            db_engine = sa.create_engine(f"mssql+pymssql://{str(self.__credentials[0])}:{str(self.__credentials[1])}@{self.server}/{db}", echo=echo)
        return db_engine
    
    @property
    def __credentials(self) -> tuple[str, str]:
        creds = keyring.get_credential(self.wincred, None)
        if creds is None:
            raise ValueError(f"Den angivne credential {self.wincred} kan ikke findes i windows credentials")        
        return (str(creds.username), str(creds.password))
       
if __name__ == '__main__':
    source = DBHelper(server=r"C2100306\MSSQLSERVER01")
    try:
        source.connect_engine(r"1_PROD_COPY")
        print("Succes!")
    except sa.exc.ProgrammingError as err:
        print(err)
    source.close()



from datetime import datetime
import re
import time
import lxml.etree as et
from dataclasses import dataclass
import io

import pandas as pd

@dataclass
class Regnskab:
    """
    """
    xlbr_response: io.BytesIO
    cvr: int
    start_date: str
    end_date: str
    translations: pd.DataFrame
    
    def __post_init__(self):
        self.__regnskab = et.parse(io.BytesIO(self.xlbr_response.content)).getroot()
        self.__namespaces: dict = self.__regnskab.nsmap
        self.__relevant_contexts = self.__extract_relevant_contexts
        self.__full_data = self.__load_full_data()
    
    def get_data_as_dict(self) -> dict:
        self.__result_dict = {
            "cvr": self.cvr,
            "start_date": self.__full_data["start_date"],
            "end_date": self.__full_data["end_date"]
            }
        self.__translations = self.__load_translations()
        for k in self.__translations.keys():
            result_var = self.__translations[k][0]
            self.__result_dict[k] = self.__full_data[result_var]
        self.__run_calculations()
        self.__run_validations()
        return self.__result_dict
   
    def __run_validations(self) -> None:
        """TODO: Funktion som validerer de beregnede værdier."""
        # Mimic beregningerne omkring bruttotest/bruttoresultat som ligger i linje 825+
        pass
    
    def __run_calculations(self) -> None:
        """Kører beregningslogik hvor tallene fra regnskabet ikke bare hentes én-til-én fra xlbr-data, men opkodes på den ene eller anden måde. 
        Oftest et spørgsmål om at centrale termer enten findes én-til-én eller opsplittet i regnskabsdata. 
        Overvej at splitte de enkelte beregninger ud i særskilte filer og dokumenter dem grundigt!"""
        self.__result_dict["afskrivninger"] = self.__afskrivninger
        self.__result_dict["vareBeholdning"] = self.__calculate_stat("varebeholdning_1", "vareBeholdning")
        self.__result_dict["anlaegsAktiver"] = self.__calculate_stat("anlaegsAktiver_1", "anlaegsAktiver")
        self.__result_dict["tilgodehavender"] = self.__calculate_stat("tilgodehavender_1", "tilgodehavender")
        self.__result_dict["personaleomkostninger"] = self.__calculate_stat("personaleOmkostninger_1", "personaleOmkostninger1")
        self.__result_dict["omsaetningsAktiver"] = self.__omsaetningsaktiver
        self.__result_dict["samletGaeld"] = self.__samlet_gaeld
        self.__result_dict["samletGaeld_test"] = self.__samlet_gaeld_test
        self.__result_dict["hensaettelser"] = self.__hensaettelser
        self.__result_dict["balance_test"] = self.__balance_test
        self.__result_dict["EBIT"] = self.__ebit
        self.__result_dict["ebitda_beregnet"] = self.__EBITDA
        self.__result_dict["brutto_tester"] = self.__brutto_test
        # self.__result_dict["ebit_test"] = self.__ebit_temp
        
    @property
    def __ebit(self) -> float:
        if "EBIT" in self.__result_dict:
            return self.__result_dict["EBIT"]
        return self.__ebit_temp

    @property
    def __EBITDA(self) -> float:
        if "EBITDA1" in self.__result_dict and self.__result_dict["EBITDA1"]:
            return self.__result_dict["EBITDA1"]
        if "EBIT" in self.__result_dict and self.__result_dict["EBIT"] and self.__result_dict['afskrivninger']:
            return self.__result_dict["EBIT"] + abs(self.__result_dict["afskrivninger"])
        return self.__result_dict["EBIT"]

    @property
    def __balance_test(self) -> float:
        hensaettelser = self.__result_dict["hensaettelser"] if self.__result_dict["hensaettelser"] else 0
        if "egenkapital" not in self.__result_dict:
            return None
        try:
            return sum([self.__samlet_gaeld_test, self.__result_dict["egenkapital"], hensaettelser])
        except TypeError:
            # print(self.cvr)
            return None

    @property
    def __hensaettelser(self) -> float:
        if "hensaettelser" in self.__result_dict:
            return self.__result_dict["hensaettelser"]
        if "hensaettelserSkat" in self.__result_dict:
            return self.__result_dict["hensaettelserSkat"]
        return None

    @property
    def __samlet_gaeld_status(self) -> bool:
        """
        Inkluderer lige nu ikke en grænse for hvor store afvigelser der accepteres.
        Hvis dette skal inkluderes, så kør noget i stil med abs(self.__samlet_gaeld-self.__samlet_gaeld_test) <= 2
        """
        return self.__samlet_gaeld == self.__samlet_gaeld_test

    @property
    def __samlet_gaeld(self) -> float:
        samlet_gaeld_vars = self.__indikator_variable("samletGaeld", recoded=True)
        if not samlet_gaeld_vars:
            try:
                return self.__result_dict["samletGaeld"]
            except KeyError:
                pass
        return sum([self.__result_dict[res] for res in samlet_gaeld_vars if res in self.__result_dict])

    @property
    def __samlet_gaeld_test(self) -> float:
        samlet_gaeld_test_vars = self.__indikator_variable("tjekGaeld", recoded=True)
        if not samlet_gaeld_test_vars:
            return None
        return sum([self.__result_dict[res] for res in samlet_gaeld_test_vars if res in self.__result_dict])

    @property
    def __ebit_temp(self) -> float:
        # EBITDA beregnes udelukkende hvis bruttoresultatet er angivet og bruttoresultat kan beregnes (__brutto_test) Ewwww
        if "bruttonResultat" not in self.__result_dict or not self.__result_dict["bruttoResultat"] or self.__result_dict["bruttoResultat"] != self.__brutto_test:
            return None
    
        resultatUdg_EBIT_beregning = ['afskrivninger_2', 'personaleomkostninger', 'distributionsOmkostninger', 'administrationsOmkostninger', 'andreDriftsOmkostninger']
    
        udgifter_foer_EBIT = sum([self.__result_dict[res] for res in resultatUdg_EBIT_beregning if res in self.__result_dict and self.__result_dict[res]])
        
        return self.__brutto_test - udgifter_foer_EBIT

    @property
    def __ebitda_unused(self) -> float:
        """Bruges ikke..."""

        self.__result_dict["EBIT"] = self.__ebit_temp if not self.__result_dict["EBIT"] else self.__result_dict["EBIT"]
        
        if not self.__result_dict["EBIT"]:
            return None
        
        if "afskrivninger_2" not in self.__result_dict:
            return self.__result_dict["EBIT"]
                
        return self.__result_dict["EBIT"] + self.__result_dict["afskrivninger_2"]

    @property
    def __brutto_test(self) -> float:
        """Mellemregning som tjekker hvorvidt bruttoResultat kan beregnes via andre parametre i regnskabet. Bruges så
        vidt vides til at afgøre om andre dele er fuld oplyst og dermed kan beregnes, ex. ebitda."""
        resultatInd = self.__indikator_variable("resultatInd", recoded=True)
        resultatUdg_1 = self.__indikator_variable("resultatUdg", recoded=True)
        # resultatUdg_2 = self.__indikator_variable("resultatUdg2", recoded=True)  # Udkommenteret indtil videre - synes ikke at blive anvendt.
        
        brutto_test = self.__result_dict["efterSkat"] if self.__result_dict["efterSkat"] else 0
        resultatInd_sum = sum([-self.__result_dict[res] for res in resultatInd if res in self.__result_dict])
        resultatUdg_sum = sum([self.__result_dict[res] for res in resultatUdg_1 if res in self.__result_dict and self.__result_dict[res]])
        
        return sum([brutto_test, resultatInd_sum, resultatUdg_sum])
    
    def __calculate_stat(self, stat: str, indikator: str) -> float:
        """Generel funktion, som på baggrund af et parameter returnerer enten parameteret, hvis dette findes én-til-én i regnskabet - alternativt en summeret indikator"""
        if stat in self.__result_dict:
            return self.__result_dict[stat]

        indikator_variable = self.__indikator_variable(indikator)
        
        if not indikator_variable:
            return None
            
        return sum([self.__full_data[self.__translations[var][0]] for var in indikator_variable if var in self.__full_data.keys()])
    
    @property
    def __omsaetningsaktiver(self) -> float:
        """Beregninger omsætningsaktiver. Returnerer xlbr stat hvis denne er angivet ellers forsøges den beregnet via to forskelligemodeller"""
        if "omsaetningsAktiver_1" in self.__result_dict:
            return self.__result_dict["omsaetningsAktiver_1"]
        
        omsakt_vars = ["vareBeholdning", "tilgodehavender", "likvider"]
        if all(k in self.__result_dict for k in omsakt_vars):
            return sum([self.__result_dict[var] for var in omsakt_vars if self.__result_dict[var] is not None])

        if all(k in self.__result_dict for k in ["aktiver", "anlaegsAktiver"]):
            omsaetningsaktiver1 = self.__result_dict["aktiver"] - 0 if not self.__result_dict["anlaegsAktiver"] else self.__result_dict["anlaegsAktiver"]
            rel_vars = ['varebeholdning', 'tilgodehavender', 'likvider', 'vaerdiPapirer']
            omsaetningsaktiver2 = sum([self.__result_dict[var] for var in rel_vars if var in self.__result_dict and self.__result_dict[var]])
            if abs(omsaetningsaktiver1-omsaetningsaktiver2) < 2:
                return omsaetningsaktiver2
            
        return None

    @property
    def __afskrivninger(self) -> float:
        """Funktion som beregner afskrivninger. Default return None - bør ikke nås. Tjek af dette indlejres i validations."""
        afskrivninger2 = self.__result_dict.get("afskrivninger_2", 0)
        
        if afskrivninger2 != 0:
            return afskrivninger2
        
        afskrivninger1 = self.__result_dict.get("afskrivninger_1", 0)
        tilbagefoerselAfAfskrivninger = self.__result_dict.get("tilbagefoerselAfAfskrivninger", 0)
        
        if afskrivninger1 != 0:
            return afskrivninger1 + tilbagefoerselAfAfskrivninger
            
        return None # Default state - denne bør ikke nås

    def __load_full_data(self) -> dict:
        """Danner en dict med samtlige resultater, som indeholdes i det downloadede xlbr regnskab"""
        full_results = {}
        full_results["cvr"] = self.cvr
        full_results["start_date"] = self.start_date
        full_results["end_date"] = self.end_date

        for elem in self.__relevant_tags:
            elem_name = re.sub(r"(\{.*\})", "", elem.tag)
            try:
                elem_value = float(elem.text)
            except TypeError:
                """Bypasser termer som er angivet i regnskabet, men uden angivelse"""
                continue
            except ValueError:
                elem_value = elem.text
            full_results[elem_name] = elem_value        
        
        return full_results
    
    def __indikator_variable(self, indikator: str, recoded: bool = False) -> float:
        return_column = "translated" if not recoded else "recoded_name"
        indikator_variable = self.__trans_df.loc[self.__trans_df[indikator] == 1, return_column].to_list()
        indikator_variable_renset = [var for var in indikator_variable if var in self.__result_dict]
        return list(set(indikator_variable_renset))
    
    def __load_translations(self) -> dict:
        self.__trans_df = self.translations
        trans_keys = self.__trans_df.translated.to_list()
        trans_values = self.__trans_df.key.to_list()
        full_data_keys = self.__full_data.keys()
       
        trans_dict = {}
        for k, v in zip(trans_keys, trans_values):
            if v not in full_data_keys:
                continue
            if k in trans_dict:
                trans_dict[k].append(v)
            else:
                trans_dict[k] = [v]   
        return trans_dict
   
    @property
    def __extract_relevant_contexts(self):
        """Funktion som udleder de relevate context id'er fra regnskabet og dermed hvilke parametre, som skal inkluderes.
        Lav løsning som laver et fallback på contextids, hvis der ikke er contexts uden "explitcitMember angivet. Se CVR øverst."""
        context_ids = []

        context_tags = [elem for elem in self.__regnskab.findall(r"{http://www.xbrl.org/2003/instance}context", self.__namespaces)]
        for context in context_tags:
            elem_names = [re.sub(r"(\{.*\})", "", elem.tag) for elem in context.iter()]
            if any(elem_name in ["explicitMember"] for elem_name in elem_names):
                    continue
            for elem in context.iter():
                if elem.text == self.end_date:
                    context_ids.append(context.get("id"))
                    continue
                if elem.text and elem.text.startswith(self.end_date[:4]):
                    # Denne håndtering er ikke umiddelbart ikke optimal. Er nødvendig da der i nogle tilfælde ikke kan testes på den fulde dato
                    # da afslutningsdatoen ikke fremgår af den enkelte context beskrivelse men datoen for godkendelse af regnskabet
                    # kan potentielt gøres mere robust ved at teste på startdatoen. 
                    context_ids.append(context.get("id"))
        # Nedenstående laver et fallback, hvis regnskabet ikke indeholder contexts, hvor "explicitMember" ikke er angivet. I EIFO scriptet håndteres dette
        # via koncern parameteret på get_regnskab() - hvis parameteret sættes til true, så anvendes explicitMember contexts. Nedenstående skal muligvis
        # udbygges så der angives en bit på det udtrukne regnskab ift. hvilken "model" for contexts, som er anvendt.
        # Er udbygget med et tjek af hvor der kun identificeres ét context id, hvorved listen udbygges. Dette er ikke en hensigtsmæssig håndtering!
        if not context_ids or len(list(set(context_ids))) < 2:
            context_tags = [elem for elem in self.__regnskab.findall(r"{http://www.xbrl.org/2003/instance}context", self.__namespaces)]
            for context in context_tags:
                elem_names = [re.sub(r"(\{.*\})", "", elem.tag) for elem in context.iter()]    
                for elem in context.iter():
                    if elem.text == self.end_date:
                        context_ids.append(context.get("id"))

        assert context_ids, f"Der blev ikke fundet context ids, hvilket er en fejl! CVR: {self.cvr}, slutdato: {self.end_date}"
        return context_ids
    
    @property
    def __relevant_tags(self):
        """Finder alle de tags i regnskabet med id'er fra de relevante contexter."""
        return [elem for elem in self.__regnskab.findall("{*}*", self.__namespaces) if elem.get("contextRef") in self.__relevant_contexts]


from dataclasses import dataclass


@dataclass
class CVROwner:
    response: dict
    cvr: int
    
    def __post_init__(self):
        self.__har_reelle_ejere = self.__extract_har_reelle_ejere()
    
    def get_data(self) -> dict | list[dict]:
        """Funktion som henter reelle ejere for et givent cvr response"""
        if not self.__deltager_relationer:
            return {
                    "cvr" : self.cvr,
                    "api_status": "Ingen deltager relationer angivet",
            }
        result_list = []
        for person in self.__personer:
            result = self.__extract_personal_info(person)
            result_list.append(result)
        return result_list
            
    def __extract_personal_info(self, p_dict: dict) -> dict:
        temp_result = {
            "cvr": self.cvr,
            "api_status":  "Deltager relationer fundet",
            "navn": p_dict["deltager"]["navne"][0]["navn"],
            "enhedsnummer": int(p_dict["deltager"]["enhedsNummer"]),
        }
        org_info = self.__extract_org_info(p_dict["organisationer"])
        return temp_result | org_info
        
    def __extract_org_info(self, org_dict: dict) -> dict:
        roller = []
        gyldig_til_datoer = []
        for org in org_dict:
            medlemsdata = org["medlemsData"][0]["attributter"]
            for v in medlemsdata:
                roller.extend(rolle["vaerdi"].lower() for rolle in v["vaerdier"] if not rolle["periode"]["gyldigTil"])
                gyldig_til_datoer.extend([m["periode"]["gyldigTil"] for m in v["vaerdier"]])
        reel_ejer = 0
        for rolle in roller:
            if "reel ejer" in rolle:
                reel_ejer = 1
            if not self.__har_reelle_ejere:
                if rolle in ["adm. dir.", "bestyrelsesmedlem", "formand", "næstformand"]:
                    reel_ejer = 2
        aktiv = 1 if not all(gyldig_til_datoer) else 0
        return {
            "reel_ejer": reel_ejer,
            "aktiv": aktiv,
            "roller": roller
        }

    @property
    def __virksomheder(self) -> list[dict]:
        return [p for p in self.__deltager_relationer if p["deltager"]["enhedstype"]=="VIRKSOMHED"]

    @property
    def __personer(self) -> list[dict]:
        return [p for p in self.__deltager_relationer if p["deltager"]["enhedstype"]!="VIRKSOMHED" if p["deltager"]]
    
    @property
    def __deltager_relationer(self) -> list[dict]:
        return [deltager for deltager in self.response["_source"]["Vrvirksomhed"]["deltagerRelation"] if deltager["deltager"]]
    
    def __extract_har_reelle_ejere(self) -> bool:
        """Grimt"""
        for delt in self.response["_source"]["Vrvirksomhed"]["deltagerRelation"]:
            for org in delt["organisationer"]:
                if org["organisationsNavn"][0]["navn"] == "Reelle ejere":
                    return False if not org["medlemsData"] else True


import json
import pandas as pd
import requests
import keyring
from requests.auth import HTTPBasicAuth
from operator import itemgetter
from dataclasses import dataclass
from Levenshtein import ratio
from tqdm import tqdm
from concurrent.futures import ThreadPoolExecutor, as_completed
# Egne klasser
from .cvr_response_class import CVRResponse
from .cvr_regnskab_class import Regnskab
from .cvrapi_functions.name_handler import run_name_handler

@dataclass
class CVRAPI:
    """Super rodet!"""
    
    def __post_init__(self):
        self._url = self._url_virk
        self.__test_setup()
        
    def __test_setup(self):
        self.__creds = keyring.get_credential("cvr_api", None)
        if self.__creds is None:
            raise ValueError("Du skal tilføje brugernavn og password til windows credentials kaldet cvr_api.")
        test_query = {"query": {"bool": {"must": [{"term": {"Vrvirksomhed.cvrNummer": "29035695"}}]}}}
        response = self._response(test_query)
        response.raise_for_status()
        print("Forbindelse til CVR etableret korrekt!")
        
    @property
    def _headers(self) -> dict:
        return {"content-type": "application/json", "cache-control": "no-cache"}
    
    @property
    def _url(self) -> str:
        return self._internal_url
    
    @_url.setter
    def _url(self, url: str):
        self._internal_url = url
    
    @property
    def _url_virk(self) -> str:
        return r"http://distribution.virk.dk/cvr-permanent/virksomhed/_search"

    @property
    def _url_penh(self) -> str:
        return r"http://distribution.virk.dk/cvr-permanent/produktionsenhed/_search"
    
    @property
    def _url_regnskab(self) -> str:
        return r"http://distribution.virk.dk/offentliggoerelser/_search"
    
    def __create_query(self) -> dict:
        query = { "query": { "match": {} } }
        query["query"]["match"] = { self.query_field: {"query": self.query_term, "operator": self.query_operator} }
        query["size"] = self.mqr
        return query

    def _response(self, payload: dict) -> requests.Response:
        resp = requests.request(
            method="POST",
            url=self._url,
            data=json.dumps(payload),
            headers=self._headers,
            auth=HTTPBasicAuth(self.__creds.username, self.__creds.password)
        )
        resp.raise_for_status()
        return resp
    
    def __parse_query(self, query: dict) -> json:
        response = self._response(query)
        return json.loads(response.content)
    
    def test_query(self, query: dict, url_virk: bool = True) -> json:
        """Funktion som anvendes til at teste nye queries inden indlejring i klassen"""
        self._url = self._url_virk if url_virk else self._url_penh
        response = self._response(query)
        json_cont = json.loads(response.content)
        print(json_cont)
        return self.__parse_responses(json_cont)
    
    def __parse_responses(self, json_cont: json) -> list:
        results = []
        if json_cont["hits"]["total"] == 0:
            return results
        for hit in json_cont["hits"]["hits"]:
            # Grim, grim, grim håndtering
            if r"/virksomhed/" in self._url:
                if hit["_source"]["Vrvirksomhed"]["virksomhedMetadata"]["nyesteNavn"] is None:
                    continue
                results.append(self.__parse_response(hit))
            else:
                results.append(self.__parse_response_penhed(hit))
        return results

    def __parse_response_penhed(self, hit: dict) -> dict:
        """
        NB! Samles med __parse_respose til én CVRResponse klasse        
        Trækker relevante informationer ud fra den returnerede json-streng. Bør nok håndteres anderledes, da 
        det er forskelligt hvilke info, som reelt er relevante. 
        Lav om til en "response klasse" som kan genanvendes andre steder."""
        result = {}
        if hit["_source"]["VrproduktionsEnhed"]["produktionsEnhedMetadata"]["nyesteNavn"] is None:
            return result
        result["name"] = hit["_source"]["VrproduktionsEnhed"]["produktionsEnhedMetadata"]["nyesteNavn"]["navn"]
        result["pnr"] = hit["_source"]["VrproduktionsEnhed"]["pNummer"]
        try:
            result["cvr"] = hit["_source"]["VrproduktionsEnhed"]["virksomhedsrelation"][0]["cvrNummer"]
        except IndexError:
            print(self.query_term)
            pass
        result["status"] = hit["_source"]["VrproduktionsEnhed"]["produktionsEnhedMetadata"]["sammensatStatus"]
        result["match_score"] = hit["_score"]
        return result
    
    def __parse_response(self, hit: dict) -> dict:
        """Trækker relevante informationer ud fra den returnerede json-streng. Bør nok håndteres anderledes, da 
        det er forskelligt hvilke info, som reelt er relevante. 
        Lav om til en "response klasse" som kan genanvendes andre steder."""
        result = {}
        if hit["_source"]["Vrvirksomhed"]["virksomhedMetadata"]["nyesteNavn"] is None:
            return result
        result["name"] = hit["_source"]["Vrvirksomhed"]["virksomhedMetadata"]["nyesteNavn"]["navn"]
        result["cvr"] = hit["_source"]["Vrvirksomhed"]["cvrNummer"]
        result["status"] = hit["_source"]["Vrvirksomhed"]["virksomhedMetadata"]["sammensatStatus"]
        result["match_score"] = hit["_score"]
        return result
    
    def __query_search_step(self, opr: str, field: str) -> list:
        """Omskrives til flere funktioner..."""
        self.query_operator = opr
        self.query_field = field
        query = self.__create_query()
        result = self.__parse_query(query)
        parsed_results = self.__parse_responses(result)
        if not parsed_results:
            return []  # Det samme som False
        for idx, pr in enumerate(parsed_results):
            temp_lratio = CVRNameComparator(self.query_term, parsed_results[idx]["name"]).lratio()
            if parsed_results[idx]["status"].lower() in ["aktiv", "normal"]:
                parsed_results[idx]["aktiv"] = 1
            else:
                parsed_results[idx]["aktiv"] = 0
            parsed_results[idx]["lratio"] = temp_lratio
            parsed_results[idx]["query_operator"] = self.query_operator
            parsed_results[idx]["query_field"] = self.query_field
            # Ualmindelig uelegant håndtering
            if opr == "AND" and len(self.query_term.split(" "))> 1 and field == "Vrvirksomhed.binavne.navn":
                parsed_results[idx]["lratio"] = 1
            
        # Oprydning - fjerner alle de hits med en lratio under grænseværdi. 
        reduced_results = [pr for pr in parsed_results if pr["lratio"] >= self.lrt]
        # Hvis der er to eller flere hits med lratio=1, så fjerne lratio=1 and aktiv=0. Kan gøres mere elegant...
        lratio_aktiv = sum(i["lratio"] for i in reduced_results if i["lratio"]==1 and i["aktiv"]==1)
        lratio_inaktiv = sum(i["lratio"] for i in reduced_results if i["lratio"]==1 and i["aktiv"]==0)
        if lratio_aktiv > 0 and lratio_inaktiv > 0:
            reduced_results = [i for i in reduced_results if i["aktiv"]==1 and i["lratio"]==1]
        # Sorterer resultaterne desc på lratio
        sorted_results = sorted(reduced_results, key=itemgetter("lratio"), reverse=True)            
        return sorted_results

    def __run_search(self) -> list:
        self._url = self._url_virk
        """Logik som itererer igennem en række forskellige søgningstyper for at opnå det bedste match."""
        search_fields = [
            "Vrvirksomhed.virksomhedMetadata.nyesteNavn.navn",
            "Vrvirksomhed.binavne.navn",
        ]
        
        for opr in ["AND", "OR"]:
            for field in search_fields:
                step = self.__query_search_step(opr, field)
                if step and max([r["lratio"] for r in step])==1:
                    return step
        # Forsøg på at ramme virksomheder, hvor der af en eller anden årsag er angivet komma i navnet. Skal sikkert ikke implementeres generelt
        if "," in self.query_term:
            for i in self.query_term.split(","):
                self.query_term = i
                for opr in ["AND", "OR"]:
                    for field in search_fields:
                        step = self.__query_search_step(opr, field)
                        if step and max([r["lratio"] for r in step])==1:
                            return step
        # Ugh - omskrives
        search_fields_penhed = [
            "VrproduktionsEnhed.produktionsEnhedMetadata.nyesteNavn.navn",
            # "Vrvirksomhed.binavne.navn",
        ]        
        self._url = self._url_penh
        for opr in ["AND", "OR"]:
            for field in search_fields_penhed:
                step = self.__query_search_step(opr, field)
                if step and max([r["lratio"] for r in step])==1:
                    return step
        return step  # Fall back - sker, hvis der ikke findes præcise hits
    
    def __handle_name(self, search_term: str) -> str:
        return run_name_handler(search_term)

    def search_based_on_name(self, 
                             name: str, 
                             pnr: bool = False, 
                             max_returned_results: int = 20, 
                             max_query_results: int = 100, 
                             lratio_thres: float = 0.5
                             ) -> list:
        # Sætter parametre som klasse variable, så de kan refereres i andre funktioner uden at parse dem som argumenter.
        self.query_term = self.__handle_name(name)
        self.mrr = max_returned_results
        self.mqr = max_query_results  # Afgører hvor mange søgeresultater cvr maksimalt skal returnere
        self.pnr_search = pnr  # Afgører om der også skal søges på p-numre eller udelukkende cvr
        self.lrt = lratio_thres  # Grænseværdi for hvilke lratio værdier, som skal medtages
        exact_search_results = self.__run_search()
        return exact_search_results
    
    def __search_based_on_cvr(self, cvr: str) -> json:
        query = {"query": {"bool": {"must": [{"term": {"Vrvirksomhed.cvrNummer": cvr}}]}}}
        self._url = self._url_virk
        response = self._response(query)
        response.raise_for_status()
        return json.loads(response.content)

    def __search_based_on_cvrp(self, cvr: str) -> json:
        query = {"query": {"bool": {"must": [{"term": {"VrproduktionsEnhed.pNummer": cvr}}]}}}
        self._url = self._url_penh
        response = self._response(query)
        response.raise_for_status()
        return json.loads(response.content)
    
    def cvr_search(self, cvr: list[int]) -> list[CVRResponse]:
        if not isinstance(cvr, list):
            raise TypeError("Input skal være en liste med cvr numre")
        results = []
        for c in cvr:
            hits_from_cvr = self.__search_based_on_cvr(str(c))
            results.append(CVRResponse(hits_from_cvr["hits"]["hits"][0]))
        return results

    def cvr_search_to_df(self, cvr: list[int], latest: bool = True, measure: str = None) -> pd.DataFrame:
        """Grimt - skal skrives om. Midlertidig løsning for at inkludere historiske data for cvr-nummeret."""
        if not isinstance(cvr, list):
            raise TypeError("Input skal være en liste med cvr numre")
        print(f"Henter CVR data for {len(cvr)} cvr-numre")
        results = []
        for c in tqdm(cvr):
            hits_from_cvr = self.__search_based_on_cvr(str(c))
            if hits_from_cvr["hits"]["total"] > 0:
                if latest:
                    api_results = CVRResponse(hits_from_cvr["hits"]["hits"][0]).get_data()
                    api_results["hit_type"] = "cvr"
                    results.append(api_results)
                else:
                    api_results = CVRResponse(hits_from_cvr["hits"]["hits"][0]).get_data(all=True, unit = "aar")
                    results.extend(api_results)
            else:
                hits_pnr = self.__search_based_on_cvrp(str(c))
                if hits_pnr["hits"]["total"] > 0:
                    api_results = CVRResponse(hits_pnr["hits"]["hits"][0]).get_data()
                    api_results["hit_type"] = "pnr"
                    api_results["pnr"] = int(c)
                    results.append(api_results)
                else:
                    results.append({"cvr": int(c)})
        return pd.DataFrame.from_dict(results)
    
    def __search_financial_based_on_cvr(self, cvr: str) -> bytes:
        self._url = self._url_regnskab
        query = {"query": {"bool": {"must": [{"term": {"cvrNummer": cvr}}, { "term": { "dokumenter.dokumentMimeType": "application" } }, { "term": { "dokumenter.dokumentMimeType": "xml" } }]}}}
        response = self._response(query)
        response.raise_for_status()
        return json.loads(response.content)
    
    def cvr_financial_search_to_df(self, cvr: list[int], limit: int = 1) -> pd.DataFrame:
        if not isinstance(cvr, list):
            raise TypeError("Input skal være en liste med cvr numre")
        print(f"Henter CVR regnskabsdata for {len(cvr)} cvr-numre")        
        results = []
        for c in tqdm(cvr):
            hits = self.__search_financial_based_on_cvr(str(c))
            if hits["hits"]["total"] > 0:
                off_regnskaber = self.__extract_xblr_urls(hits)
                if off_regnskaber:
                    results_temp = self.__handle_xblr_documents(off_regnskaber, limit)
                    results.extend(results_temp)
                else:
                    results.append({"cvr": c, "tom_aarsag": "Hits, men ingen xblr regnskaber"})
                pass
            else:
                results.append({"cvr": c, "tom_aarsag": "Ingen hits"})
                
        return pd.DataFrame.from_dict(results)
    
    # def __handle_xblr_documents(self, documents: list[dict], limit: int) -> list[dict]:
    #     results = []
    #     documents_to_proces = documents[:limit]
    #     for doc in documents_to_proces:
    #         url = doc["dokument_url"]
    #         downloaded_doc = self.__download_xblr_document(url)
    #         parsed_doc = Regnskab(downloaded_doc, documents_to_proces["start_dato"], documents_to_proces["slut_dato"]).get_data_as_dict()
    #         results.append(parsed_doc)
    #     return results

    def __handle_xblr_documents(self, documents: list[dict], limit: int) -> list[dict]:
        results = []
        documents_to_process = documents[:limit]

        def process_doc(doc: dict):
            url = doc["dokument_url"]
            downloaded_doc = self.__download_xblr_document(url)
            parsed_doc = Regnskab(downloaded_doc, doc["cvr"], doc["start_dato"], doc["slut_dato"]).get_data_as_dict()
            return parsed_doc            

        with ThreadPoolExecutor(max_workers=None) as executor:  # Adjust the number of workers as needed
            # Submit tasks to the executor
            futures = [executor.submit(process_doc, doc) for doc in documents_to_process]
            
            # Collect results as they complete
            for future in as_completed(futures):
                results.append(future.result())  # Get the result from each completed future

        return results
                
    def __download_xblr_document(self, url: str) -> bytes:
        xblr_document = requests.request(
            method="GET",
            url=url,
            headers=self._headers,
            auth=HTTPBasicAuth(self.__creds.username, self.__creds.password)
            )
        xblr_document.raise_for_status()
        return xblr_document
    
    def __extract_xblr_urls(self, hits: bytes) -> list[dict]:
        regnskaber = [hit for hit in hits["hits"]["hits"] if hit["_source"]["offentliggoerelsestype"]]
        regnskaber_red = []
        for r in regnskaber:
            output = {}
            output["cvr"] = r["_source"]["cvrNummer"]            
            output["start_dato"] = r["_source"]["regnskab"]["regnskabsperiode"]["startDato"]
            output["slut_dato"] = r["_source"]["regnskab"]["regnskabsperiode"]["slutDato"]
            for dok in r["_source"]["dokumenter"]:
                if dok["dokumentMimeType"] == r"application/xml":
                    output["dokument_url"] = dok["dokumentUrl"]
            regnskaber_red.append(output)
        regnskaber_red.sort(key=lambda x: x["start_dato"], reverse=True)
        return regnskaber_red


@dataclass
class CVRNameComparator:
    """...
    t1: str   
    Input streng - navnet på virksomheden som skal tjekkes op imod output fra CVR søgning
    t2: str    
    Output streng fra cvr søgning"""
    t1: str
    t2: str
    
    def __post_init__(self):
        self.t1_cleaned, self.t2_cleaned = self.__prep_strings()
    
    def __prep_strings(self) -> list:
        temp_t1, temp_t2 = self.t1.lower().split(" "), self.t2.lower().split(" ")
        return temp_t1, temp_t2
    
    def __ratio_case_1(self) -> float:
        # Case 1 - der er i indtastningen udeladt virk.form, hvilket synes at være en relativt ofte forekommende udfordring
        virk_typer = ["a/s", "aps", "i/s", "s/i"]
        if not [o for o in self.t1_cleaned if o in virk_typer]:
            temp_t2_red = [o for o in self.t2_cleaned if o not in virk_typer]
            temp_lratio = ratio(self.t1_cleaned, temp_t2_red)
            if temp_lratio == 1:
                return temp_lratio
        return 0
   
    def __ratio_case_default(self) -> float:
        # Case 2 - generel case
        # Starter med at fjerne termer fra input og output som overlapper for på den måde udelukkende at beregne ratio på restmængden.
        temp_t1 = " ".join([i for i in self.t1_cleaned if i not in self.t2_cleaned])
        temp_t2 = " ".join([i for i in self.t2_cleaned if i not in self.t1_cleaned])
        return ratio(temp_t1, temp_t2)
    
    def lratio(self) -> float:
        temp_lratio = self.__ratio_case_1()
        if temp_lratio == 1:
            return temp_lratio
        # temp_lratio = self.__ratio_case_2()
        return self.__ratio_case_default()